/* NetworkManager-libreswan -- Network Manager Libreswan plugin
 *
 * Dan Williams <dcbw@redhat.com>
 * Avesh Agarwal <avagarwa@redhat.com>
 * Lubomir Rintel <lkundrak@v3.sk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Copyright (C) 2010 - 2015 Red Hat, Inc.
 */

#include "nm-default.h"

#include "nm-utils/nm-shared-utils.h"

#include "utils.h"

#include <unistd.h>
#include <string.h>
#include <errno.h>

static gboolean
printable_val (GString *str, const char *val, GError **error)
{
	const char *p;

	g_return_val_if_fail (val, FALSE);

	for (p = val; *p != '\0'; p++) {
		/* Printable characters except " and space allowed. */
		if (*p != '"' && !g_ascii_isspace (*p) && g_ascii_isprint (*p))
			continue;
		g_set_error (error, NM_UTILS_ERROR, NM_UTILS_ERROR_UNKNOWN,
			     _("Invalid character in '%s'"), val);
		return FALSE;
	}

	g_string_append (str, val);
	g_string_append_c (str, '\n');
	return TRUE;
}

static gboolean
string_val (GString *str, const char *val, GError **error)
{
	const char *p;

	g_return_val_if_fail (val, FALSE);

	for (p = val; *p != '\0'; p++) {
		/* Printable characters except " allowed. */
		if (*p != '"' && g_ascii_isprint (*p))
			continue;
		g_set_error (error, NM_UTILS_ERROR, NM_UTILS_ERROR_UNKNOWN,
			     _("Invalid character in '%s'"), val);
		return FALSE;
	}

	g_string_append_printf (str, "\"%s\"\n", val);
	return TRUE;
}

static inline gboolean
optional_string_val (GString *str, const char *key, const char *val, GError **error)
{
	if (val == NULL || val[0] == '\0')
		return TRUE;
	g_string_append_c (str, ' ');
	g_string_append (str, key);
	g_string_append_c (str, '=');

	if (!string_val (str, val, error)) {
		g_prefix_error (error, _("Invalid value for '%s': "), key);
		return FALSE;
	}

	return TRUE;
}

static inline gboolean
optional_printable_val (GString *str, const char *key, const char *val, GError **error)
{
	if (val == NULL || val[0] == '\0')
		return TRUE;

	g_string_append_c (str, ' ');
	g_string_append (str, key);
	g_string_append_c (str, '=');

	if (!printable_val (str, val, error)) {
		g_prefix_error (error, _("Invalid value for '%s': "), key);
		return FALSE;
	}

	return TRUE;
}


static inline gboolean
optional_printable (GString *str, NMSettingVpn *s_vpn, const char *key, GError **error)
{
	return optional_printable_val (str,
	                               key,
	                               nm_setting_vpn_get_data_item (s_vpn, key),
	                               error);
}

char *
nm_libreswan_get_ipsec_conf (NMSettingVpn *s_vpn,
                             const char *con_name,
                             const char *leftupdown_script,
                             gboolean openswan,
                             gboolean trailing_newline,
                             GError **error)
{
	GString *ipsec_conf = NULL;
	const char *username;
	const char *phase1_alg_str;
	const char *phase2_alg_str;
	const char *right;
	const char *leftid;

	g_return_val_if_fail (NM_IS_SETTING_VPN (s_vpn), FALSE);
	g_return_val_if_fail (!error || !*error, FALSE);
	g_return_val_if_fail (con_name && *con_name, FALSE);

	ipsec_conf = g_string_sized_new (1024);
	g_string_append (ipsec_conf, "conn ");
	if (!printable_val (ipsec_conf, con_name, error)) {
		g_prefix_error (error, _("Bad connection name: "));
		goto error;
	}

	if (leftupdown_script) {
		g_string_append (ipsec_conf, " auto=add\n");
		g_string_append (ipsec_conf, " nm-configured=yes\n");
		g_string_append (ipsec_conf, " leftupdown=");
		if (!string_val (ipsec_conf, leftupdown_script, error))
			g_return_val_if_reached (FALSE);
	}

	right = nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_RIGHT);
	if (right && right[0] != '\0') {
		g_string_append (ipsec_conf, " right=");
		if (!printable_val (ipsec_conf, right, error)) {
			g_prefix_error (error, _("Invalid value for '%s': "),
					NM_LIBRESWAN_RIGHT);
			goto error;
		}
	} else {
		g_set_error (error, NM_UTILS_ERROR, NM_UTILS_ERROR_UNKNOWN,
			     _("'%s' key needs to be present."), NM_LIBRESWAN_RIGHT);
		goto error;
	}

	leftid = nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_LEFTID);
	if (leftid && leftid[0] != '\0') {
		g_string_append (ipsec_conf, " aggrmode=yes\n");
		g_string_append (ipsec_conf, " leftid=@");
		if (!printable_val (ipsec_conf, leftid, error)) {
			g_prefix_error (error, _("Invalid value for '%s': "),
			                NM_LIBRESWAN_LEFTID);
			goto error;
		}
	}

	username = nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_LEFTXAUTHUSER);
	if (username == NULL || username[0] == '\0')
		username = nm_setting_vpn_get_user_name (s_vpn);
	if (username != NULL && username[0] != '\0') {
		g_string_append (ipsec_conf, " leftxauthusername=");
		if (!string_val (ipsec_conf, username, error)) {
			g_prefix_error (error, _("Invalid username: "));
			goto error;
		}
	}

	phase1_alg_str = nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_IKE);
	if (phase1_alg_str == NULL || phase1_alg_str[0] == '\0')
		phase1_alg_str = "aes-sha1";
	if (!optional_string_val (ipsec_conf, NM_LIBRESWAN_IKE, phase1_alg_str, error))
		goto error;

	phase2_alg_str = nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_ESP);
	if (phase2_alg_str == NULL || phase2_alg_str[0] == '\0')
		phase2_alg_str = "aes-sha1;modp1024";
	if (!optional_string_val (ipsec_conf, NM_LIBRESWAN_ESP, phase2_alg_str, error))
		goto error;

	if (!openswan && g_strcmp0 (nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_VENDOR), "Cisco") == 0)
		g_string_append (ipsec_conf, " cisco-unity=yes\n");

	if (!optional_printable (ipsec_conf, s_vpn, NM_LIBRESWAN_DPDTIMEOUT, error))
		goto error;

	g_string_append (ipsec_conf, " authby=secret\n");
	g_string_append (ipsec_conf, " left=%defaultroute\n");
	g_string_append (ipsec_conf, " leftxauthclient=yes\n");
	g_string_append (ipsec_conf, " leftmodecfgclient=yes\n");
	g_string_append (ipsec_conf, " rightsubnet=0.0.0.0/0\n");
	g_string_append (ipsec_conf, " rekey=yes\n");
	g_string_append (ipsec_conf, " salifetime=24h\n");
	g_string_append (ipsec_conf, " ikelifetime=24h\n");
	g_string_append (ipsec_conf, " keyingtries=1\n");
	g_string_append (ipsec_conf, " remote_peer_type=cisco\n");
	g_string_append (ipsec_conf, " rightxauthserver=yes\n");
	g_string_append (ipsec_conf, " rightmodecfgserver=yes\n");
	g_string_append (ipsec_conf, " modecfgpull=yes");
	if (trailing_newline)
		g_string_append_c (ipsec_conf, '\n');

	return g_string_free (g_steal_pointer (&ipsec_conf), FALSE);
error:
	g_string_free (ipsec_conf, TRUE);
	return FALSE;
}
